package ru.dbotthepony.mc.otm.capability.matter

import net.minecraftforge.common.capabilities.ICapabilityProvider
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.FlowDirection
import ru.dbotthepony.mc.otm.core.Decimal
import ru.dbotthepony.mc.otm.core.RGBAColor
import ru.dbotthepony.mc.otm.core.orNull
import kotlin.math.roundToInt

interface IMatterHandler {
	/**
	 * If this is false, then [storedMatter] will throw [UnsupportedOperationException] when trying to set it
	 */
	val canSetMatterLevel: Boolean get() = true

	/**
	 * How much matter (estimated) is stored in this object. Why estimated? Because some objects can be bottomless.
	 *
	 * Implementations are free to throw [UnsupportedOperationException] if setting battery level is not supported
	 * due to technical complications (in this case, [canSetMatterLevel] MUST be false)
	 *
	 * @throws [UnsupportedOperationException]
	 */
	var storedMatter: Decimal

	/**
	 * How much matter (estimated) can this object hold. Why estimated? Because some objects can be bottomless.
	 *
	 * **DO NOT use this to determine "how much stuff is missing", use [missingMatter] instead!**
	 */
	val maxStoredMatter: Decimal

	/**
	 * Empties matter stored of this matter storage
	 *
	 * @throws [UnsupportedOperationException]
	 * @see storedMatter
	 */
	fun emptyMatter() {
		storedMatter = Decimal.ZERO
	}

	/**
	 * Fully fills matter stored in this matter storage
	 *
	 * @throws [UnsupportedOperationException]
	 * @see storedMatter
	 */
	fun fillMatter() {
		storedMatter = maxStoredMatter
	}

	/**
	 * Make this object receive matter, from outside world.
	 *
	 * Call this if you don't own the object in question.
	 *
	 * @return matter accepted
	 */
	fun receiveMatterOuter(howMuch: Decimal, simulate: Boolean): Decimal

	/**
	 * Make this object receive matter, from inside world.
	 *
	 * Call this if you DO own the object in question.
	 *
	 * Nothing stops you from calling this from outside world, however, this will most likely
	 * lead to bugs, use [receiveMatterOuter] in this case instead. Methods were separated (and inside interface) for ease of use.
	 *
	 * Don't try to hammer nails with a screwdriver.
	 *
	 * @return matter accepted
	 */
	fun receiveMatterInner(howMuch: Decimal, simulate: Boolean): Decimal

	/**
	 * Make this object extract matter from itself, from outside world.
	 *
	 * Call this if you don't own the object in question.
	 *
	 * @return matter extracted
	 */
	fun extractMatterOuter(howMuch: Decimal, simulate: Boolean): Decimal

	/**
	 * Make this object extract matter from itself, from inside world.
	 *
	 * Call this if you DO own the object in question.
	 *
	 * Nothing stops you from calling this from outside world, however, this will most likely
	 * lead to bugs, use [extractMatterOuter] in this case instead. Methods were separated (and inside interface) for ease of use.
	 *
	 * Don't try to hammer nails with a screwdriver.
	 *
	 * @return matter extracted
	 */
	fun extractMatterInner(howMuch: Decimal, simulate: Boolean): Decimal

	/**
	 * How much matter (estimated) is missing in this object. Why estimated? Because some objects can be bottomless.
	 *
	 * Use this to determine whenever you need an estimate on how much matter this object can accept, **and do not implement [maxStoredMatter] - [storedMatter] logic by yourself**.
	 */
	val missingMatter: Decimal
		get() = maxStoredMatter.minus(storedMatter).moreThanZero()

	/**
	 * Which direction does matter flows
	 */
	val matterDirection: FlowDirection
}

inline val IMatterHandler.canExtractMatter: Boolean
	get() = matterDirection.output
inline val IMatterHandler.canReceiveMatter: Boolean
	get() = matterDirection.input

fun IMatterHandler.getBarWidth(): Int {
	return ((storedMatter / maxStoredMatter).toFloat().coerceAtLeast(0f).coerceAtMost(1f) * 13f).roundToInt()
}

fun IMatterHandler.getBarColor(): Int {
	return RGBAColor.LOW_MATTER.linearInterpolation((storedMatter / maxStoredMatter).toFloat(), RGBAColor.FULL_MATTER).toInt()
}

val ICapabilityProvider.matter: IMatterHandler? get() = getCapability(MatteryCapability.MATTER).orNull()
